import _ol_ from '../index';
import _ol_layer_Layer_ from '../layer/layer';
import _ol_renderer_Type_ from '../renderer/type';
import _ol_renderer_canvas_ImageLayer_ from '../renderer/canvas/imagelayer';
import _ol_renderer_webgl_ImageLayer_ from '../renderer/webgl/imagelayer';

/**
 * @classdesc
 * Server-rendered images that are available for arbitrary extents and
 * resolutions.
 * Note that any property set in the options is set as a {@link ol.Object}
 * property on the layer object; for example, setting `title: 'My Title'` in the
 * options means that `title` is observable, and has get/set accessors.
 *
 * @constructor
 * @extends {ol.layer.Layer}
 * @fires ol.render.Event
 * @param {olx.layer.ImageOptions=} opt_options Layer options.
 * @api
 */
var _ol_layer_Image_ = function(opt_options) {
  var options = opt_options ? opt_options : {};
  _ol_layer_Layer_.call(this,  /** @type {olx.layer.LayerOptions} */ (options));
};

_ol_.inherits(_ol_layer_Image_, _ol_layer_Layer_);


/**
 * @inheritDoc
 */
_ol_layer_Image_.prototype.createRenderer = function(mapRenderer) {
  var renderer = null;
  var type = mapRenderer.getType();
  if (_ol_.ENABLE_CANVAS && type === _ol_renderer_Type_.CANVAS) {
    renderer = new _ol_renderer_canvas_ImageLayer_(this);
  } else if (false && type === _ol_renderer_Type_.WEBGL) {
    renderer = new _ol_renderer_webgl_ImageLayer_((mapRenderer), this);
  }
  return renderer;
};


/**
 * Return the associated {@link ol.source.Image source} of the image layer.
 * @function
 * @return {ol.source.Image} Source.
 * @api
 */
_ol_layer_Image_.prototype.getSource;
export default _ol_layer_Image_;
